unit Evaluator;

{
  Unit ten pozwala interpretowa i oblicza rwnania
  Autor - Pixel (markovcd@gmail.com)                         
  http://4programmers.net/Delphi/Gotowce/Interpreter_rwna
  rdo: http://pl.wikipedia.org/wiki/ONP

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>.
}

interface

uses
  SysUtils, Classes, Math;

const
  // mozna dowolnie zmieniac typ nawiasow, o ile nie koliduja z innymi symbolami
  LEFT_PARENTHESIS = '(';
  RIGHT_PARENTHESIS = ')';
  SYMBOL_ERROR = 'Nieznany symbol: "%s".';
  PARENTHESIS_ERROR = 'le umieszczone nawiasy.';
  NAME_ERROR = 'Niedozwolona nazwa: "%s".';
  NAME_ERROR2 = 'Nazwa nie moe zaczyna si od cyfry.';
  OPERATOR_ERROR = 'Operator nie moe zawiera cyfr.';

type
  // typy operacji odpowiednio dla funkcji i operatorow
  TMathFunction1 = function(const a: Extended): Extended;
  TMathFunction2 = function(const a, b: Extended): Extended;

  // rekord zawiera wlasciwosci wszystkich rozpoznawanych symboli
  TSymbol = record
    Name: string[255];
    case Integer of
      0: ( // operator
        AssociativeRight: Boolean; // czy prawostronnie laczny
        Priority: Byte; // im wyzsza wartosc tym wazniejszy operator
        Operation2: TMathFunction2);
      1: ( // funckcja
        Operation1: TMathFunction1);
      2: ( // stala
        Value: Extended);
  end;

  // stos
  TStringStack = class(TStringList)
  public
    function Pop: string; // zdejmuje element ze stosu
    function Peek: string; // podejrzenie ostatniego elementu
    procedure Push(Value: string); // odklada element na stos
    function IsEmpty: Boolean; // czy stos pusty?
    procedure DeleteEmpty; // usuwanie pustych elementow
  end;

  // lista zawierajaca symbole rozpoznawane przez klase TEval
  TSymbolList = class abstract(TObject)
  private
    FItems: array of TSymbol;
    FCaseSensitive: Boolean;
  protected
    procedure IsAlfanumeric(Value: string);
    function GetItem(Name: string; var Symbol: TSymbol): Integer; virtual;
  public
    procedure AddItem(Item: TSymbol; SafeAdd: Boolean = True); // dodaje element (SafeAdd = True nadpisuje stary element)
    procedure Delete(Index: Integer); // usuwa element
    function Items(Index: Integer): TSymbol; // tablica z elementami
    procedure Clear; // wyczyszczenie zawartosci tablicy
    function Count: Integer; // zwrocenie ilosci elementow
    property CaseSensitive: Boolean read FCaseSensitive write FCaseSensitive default False; // wrazliwosc na rozmiar liter
    function Find(Name: string): TSymbol; // odszukanie elementu po nazwie
    function IsMember(Name: string): Boolean; // sprawdzenie czy element nalezy do listy
    function IsMember2(Name: string): Boolean; // to samo co wyzej, ale uwzglednia minus przed symbolem
    function GetIndex(Name: string): Integer; // pobranie indeksu elementu po nazwie
  end;

  TOperators = class(TSymbolList)
  public
    procedure Add(Name: string; Operation: TMathFunction2; Priority: Integer;
      AssociativeRight: Boolean; SafeAdd: Boolean = False); // dodaje element, wygodna forma
  end;

  TFunctions = class(TSymbolList)
  public
    procedure Add(Name: string; Operation: TMathFunction1; SafeAdd: Boolean = False); // dodaje element, wygodna forma
  end;

  TConstants = class(TSymbolList)
  public
    procedure Add(Name: string; Value: Extended; SafeAdd: Boolean = False); // dodaje element, wygodna forma
  end;

  // glowna klasa
  TEval = class(TObject)
  private
    FOperators: TOperators;
    FFUnctions: TFunctions;
    FConstants: TConstants;
    FExpression: string;
    FPostfixExpression: string;
    FResult: string;
   protected
    procedure WriteExpression(Value: string);
    procedure WritePostfixExpression(Value: string);
    function IsNumber(Value: string): Boolean;
    function IsNumber2(Value: string): Boolean;
  public
    constructor Create(DefaultSet: Boolean = True); // True jezeli twozyc domyslne symbole
    destructor Destroy; override;
    procedure Default; // przywrocenie wszystkiego do stanu domyslnego
    function PostfixEval(Value: string): string;
    function InfixToPostfix(Value: string): string;
    function PostfixToInfix(Value: string): string;
    property Operators: TOperators read FOperators write FOperators; // lista operatorow
    property Functions: TFunctions read FFunctions write FFunctions; // lista funkcji
    property Constants: TConstants read FConstants write FConstants; // lista stalych
    property Expression: string read FExpression write WriteExpression; // przypisz wyr. matematyczne aby go obliczyc
    property PostfixExpression: string read FPostfixExpression write WritePostfixExpression; // odczytanie wyr. w formacie ONP
    property EvalResult: string read FResult; // zwraca wynik dzialania
  end;

// funkcje pomoga zdefiniowac podstawowe symbole:
function Plus(const a, b: Extended): Extended; // dodawania
function Substract(const a, b: Extended): Extended; // odejmowania
function Multiply(const a, b: Extended): Extended; // mnozenia
function Divide(const a, b: Extended): Extended; // dzielenia
function Abs(const a: Extended): Extended; // wartosci bezwzglednej

implementation

function Plus(const a, b: Extended): Extended;
begin
  Result := a + b;
end;

function Substract(const a, b: Extended): Extended;
begin
  Result := a - b;
end;

function Multiply(const a, b: Extended): Extended;
begin
  Result := a * b;
end;

function Divide(const a, b: Extended): Extended;
begin
  Result := a / b;
end;

function Abs(const a: Extended): Extended;
begin
  if a < 0 then Result := -a
  else Result := a;
end;

// -------------- TStringStack -------------- //
function TStringStack.Pop: string;
begin
  Result := Peek;
  Delete(Count - 1);
end;

function TStringStack.Peek: string;
begin
  Result := Strings[Count - 1];
end;

procedure TStringStack.Push(Value: string);
begin
  Add(Value);
end;

function TStringStack.IsEmpty: Boolean;
begin
  Result := (Count = 0);
end;

procedure TStringStack.DeleteEmpty;
var
  i: Integer;
begin
  for i := Count - 1 downto 0 do
    if Trim(Strings[i]) = '' then Delete(i);
end;
// -------------- TStringStack -------------- //

// -------------- TSymbolList --------------- //
procedure TSymbolList.AddItem(Item: TSymbol; SafeAdd: Boolean = True);
var
  i: Integer;
begin
  if SafeAdd then begin
    i := GetIndex(Item.Name);
    if i <> -1 then Delete(i);
  end;
  SetLength(FItems, High(FItems) + 2);
  FItems[High(FItems)] := Item;
end;

function TSymbolList.Items(Index: Integer): TSymbol;
begin
  Result := FItems[Index];
end;

procedure TSymbolList.Clear;
begin
  SetLength(FItems, 0);
end;

function TSymbolList.Count: Integer;
begin
  Result := Length(FItems);
end;

procedure TSymbolList.Delete(Index: Integer);
var
  ALength: Integer;
  i: Cardinal;
begin
  ALength := Length(FItems);
  Assert(ALength > 0);
  Assert(Index < ALength);
  for i := Index + 1 to ALength - 1 do
    FItems[i - 1] := FItems[i];
  SetLength(FItems, ALength - 1);
end;

function TSymbolList.GetItem(Name: string; var Symbol: TSymbol): Integer;
var
  i: Integer;
  s: string;
begin
  Result := -1;
  if not FCaseSensitive then
    Name := LowerCase(Name);
  for i := 0 to High(FItems) do begin
    s := FItems[i].Name;
    if not FCaseSensitive then
      s := LowerCase(s);
    if s = Name then begin
      Symbol := FItems[i];
      Result := i;
      Break;
    end;
  end;
end;

function TSymbolList.Find(Name: string): TSymbol;
var
  s: TSymbol;
begin
  GetItem(Name, s);
  Result := s;
end;

function TSymbolList.GetIndex(Name: string): Integer;
var
  s: TSymbol;
begin
  Result := GetItem(Name, s);
end;

function TSymbolList.IsMember(Name: string): Boolean;

begin
  Result := (GetIndex(Name) <> -1);
end;

function TSymbolList.IsMember2(Name: string): Boolean;
var
  tmp: string;
begin
  if Name[1] = '-' then tmp := Copy(Name, 2, Length(Name))
  else tmp := Name;
  Result := IsMember(tmp);
end;

procedure TSymbolList.IsAlfanumeric(Value: string);
var
  i: Integer;
begin
  if Value[1] in ['0'..'9'] then
    raise Exception.Create(NAME_ERROR2);

  for i := 1 to Length(Value) do
    if not (Value[i] in ['a'..'z', 'A'..'Z', '0'..'9']) then
      raise Exception.Create(Format(NAME_ERROR, [Value]));
end;
// -------------- TSymbolList --------------- //

// --------------- TOperators --------------- //
procedure TOperators.Add(Name: string; Operation: TMathFunction2; Priority: Integer;
  AssociativeRight: Boolean; SafeAdd: Boolean = False);
var
  Item: TSymbol;
  i: Integer;
begin
  for i := 1 to Length(Name) do
    if (Name[i] in ['0'..'9']) then
      raise Exception.Create(OPERATOR_ERROR);

  Item.Name := Name;
  Item.Operation2 := Operation;
  Item.Priority := Priority;
  Item.AssociativeRight := AssociativeRight;
  AddItem(Item, SafeAdd);
end;
// --------------- TOperators --------------- //

// --------------- TFunctions --------------- //
procedure TFunctions.Add(Name: string; Operation: TMathFunction1; SafeAdd: Boolean = False);
var
  Item: TSymbol;
begin
  IsAlfanumeric(Name);
  Item.Name := Name;
  Item.Operation1 := Operation;
  AddItem(Item, SafeAdd);
end;
// --------------- TFunctions --------------- //

// --------------- TConstants --------------- //
procedure TConstants.Add(Name: string; Value: Extended; SafeAdd: Boolean = False);
var
  Item: TSymbol;
begin
  IsAlfanumeric(Name);
  Item.Name := Name;
  Item.Value := Value;
  AddItem(Item, SafeAdd);
end;
// --------------- TConstants --------------- //

// ----------------- TEval ------------------ //
constructor TEval.Create(DefaultSet: Boolean = True);
begin
  inherited Create;
  FOperators := TOperators.Create;
  FFunctions := TFunctions.Create;
  FConstants := TConstants.Create;
  if DefaultSet then Default
  else begin // te operatory sa niezbedne do dzialania klaasy
    FOperators.Add('-', Substract, 0, False);
    FOperators.Add('*', Multiply, 1, False);
  end;
end;

destructor TEval.Destroy;
begin
  FOperators.Free;
  FFunctions.Free;
  FConstants.Free;
  inherited;
end;

function TEval.IsNumber(Value: string): Boolean;
begin
  if Value[1] = '-' then Result := False
  else Result := IsNumber2(Value);
end;

function TEval.IsNumber2(Value: string): Boolean;
var
  s: Extended;
begin
  if (Value[1] = '+') then Result := False
  else
    Result := TextToFloat(PChar(Value), s, fvExtended);
end;

procedure TEval.Default;
begin
  with FOperators do begin
    Clear;
    Add('+', Plus, 0, False);
    Add('-', Substract, 0, False);
    Add('*', Multiply, 1, False);
    Add('/', Divide, 1, False);
    Add('^', Power, 2, True);
  end;

  with FFunctions do begin
    Clear;
    Add('sqrt', Sqrt);
    Add('sin', Sin);
    Add('cos', Cos);
    Add('tan', Tan);
    Add('cot', Cot);
    Add('arcsin', ArcSin);
    Add('arccos', ArcCos);
    Add('arctan', ArcTan);
    Add('arccot', ArcCot);
    Add('sinh', Sinh);
    Add('cosh', Cosh);
    Add('tanh', Tanh);
    Add('coth', Coth);
    Add('ln', Ln);
    Add('log', Log10);
    Add('log2', Log2);
    Add('abs', Abs);
  end;

  with FConstants do begin
    Clear;
    Add('pi', Pi);
    Add('e', 2.718281828459045);
  end;
end;

procedure TEval.WriteExpression(Value: string);
begin
  FExpression := Value;
  FPostfixExpression := '';
  if Trim(FExpression) = '' then Exit;
  FPostfixExpression := InfixToPostfix(Value);
  FResult := PostfixEval(FPostfixExpression);
end;

procedure TEval.WritePostfixExpression(Value: string);
begin
  FPostfixExpression := Value;
  FExpression := '';
  if Trim(Value) = '' then Exit;
  FResult := PostfixEval(Value);
  FExpression := PostfixToInfix(Value);
end;

// funkcja liczy wartosc wyrazenia w formacie ONP
function TEval.PostfixEval(Value: string): string;
var
  i: Integer;
  input, stack: TStringStack;
  a, b, c: Extended;
  sym: TSymbol;
  tmp, tmp2: string;
begin
  try
    // tworzenie stosow
    input := TStringStack.Create;
    stack := TStringStack.Create;
    
    // wrzucenie tokenow na stos
    input.Delimiter := #32;
    input.DelimitedText := Value;
    input.DeleteEmpty;

    // zamiana staych na ich wartosci
    for i := 0 to input.Count - 1 do begin
      if input.Strings[i][1] = '-' then begin
        tmp := Copy(input.Strings[i], 2, Length(input.Strings[i]));
        tmp2 := '-';
      end else begin
        tmp := input.Strings[i];
        tmp2 := '';
      end;
      if FConstants.GetItem(tmp, sym) <> -1 then begin
        c := sym.Value;
        if tmp2 = '-' then c := -c;
        input.Strings[i] := FloatToStr(c);
      end;
    end;

    //glowna petla
    for i := 0 to input.Count - 1 do begin // Dla wszystkich symboli z wyraenia ONP wykonuj:
      tmp := input.Strings[i];
      if IsNumber2(tmp) then // jeli i-ty symbol jest liczb,
        stack.Push(tmp) // to od go na stos,
      else if FFunctions.IsMember2(tmp) then begin // jeli i-ty symbol jest funkcj to:
        a := StrToFloat(stack.Pop); // zdejmij ze stosu oczekiwan ilo parametrw (obsluga tylko jednego parametru)
        if tmp[1] = '-' then begin
          tmp := Copy(tmp, 2, Length(tmp));
          tmp2 := '-';
        end else tmp2 := '';
        sym := FFunctions.Find(tmp);
        c := sym.Operation1(a);
        if tmp2 = '-' then c := -c;
        stack.Push(FloatToStr(c)); // od na stos wynik funkcji
      end else if FOperators.GetItem(tmp, sym) <> -1 then begin // jeli i-ty symbol jest operatorem to:
        a := StrToFloat(stack.Pop); // zdejmij ze stosu jeden element (ozn. a),
        b := StrToFloat(stack.Pop); // zdejmij ze stosu kolejny element (ozn. b),
        c := sym.Operation2(b, a);
        stack.Push(FloatToStr(c)); // od na stos warto b operator a.
      end else // nieznany symbol, wymus blad
        raise Exception.Create(Format(SYMBOL_ERROR, [input.Strings[i]]));
    end;

    stack.Delimiter := #32;
    Result := stack.DelimitedText; // Zdejmij ze stosu wynik.
  finally
    // zwalnianie stosow
    input.Free;
    stack.Free;
  end;
end;

// funkcja konwertuje wyrazenie na format ONP
function TEval.InfixToPostfix(Value: string): string;
var
  i, j: Integer;
  input, stack, output: TStringStack;
  o1, o2, sym: TSymbol;
  tmp, tmp2: string;
  Error, b: Boolean;
begin
  try
    Error := False;
    Value := StringReplace(Value, ' ', '', [rfReplaceAll]); // pozbycie sie spacji

    // tworzenie stosow
    input := TStringStack.Create;
    stack := TStringStack.Create;
    output := TStringStack.Create;

    // odkladanie kazdego tokena na stos
    i := 1;
    while i <= Length(Value) do begin
      j := Length(Value) - i + 1;
      while j >= 1 do begin
        tmp := Copy(Value, i, j);
        if IsNumber(tmp)            or // jezeli numer lub
           FFunctions.IsMember(tmp) or // funkcja lub
           FConstants.IsMember(tmp) or // stala lub
           FOperators.IsMember(tmp) or // operator lub
           (tmp = LEFT_PARENTHESIS) or (tmp = RIGHT_PARENTHESIS) then begin // nawiasy to
          i := i + Length(tmp) - 1; // przesun petle o dlugosc wyrazenia
          input.Push(tmp); // dodaj na wejscie
          Break;
        end;
        Dec(j);
        if j = 0 then Error := True; // nie znaleziono symbolu!
      end;
      Inc(i);
    end;
    input.DeleteEmpty;

    if Error then begin // obsluga bledow
      tmp := StringReplace(input.Text, #13#10, '', [rfReplaceAll]);
      if Trim(tmp) <> '' then begin
        Error := False;
        for i := 1 to Length(Value) do
          if Value[i] <> tmp[i] then begin
            Error := True;
            tmp2 := tmp2 + Value[i];
            Insert(' ', tmp, i);
          end else if Error then Break;
      end else tmp2 := Value;
      raise Exception.Create(Format(SYMBOL_ERROR, [tmp2]));
    end;

    // obsluga ujemnych liczb, stalych, funkcji
    for i := input.Count - 1 downto 1 do begin
      if (input.Strings[i] = '-') and (FOperators.IsMember(input.Strings[i - 1]) or (input.Strings[i - 1] = LEFT_PARENTHESIS)) then begin
        if input.Strings[i + 1] <> LEFT_PARENTHESIS then begin
          input.Strings[i] := '-' + input.Strings[i + 1];
          input.Delete(i + 1);
        end else begin
          input.Strings[i] := '-1';
          input.Insert(i + 1, '*');
        end;
      end;
    end;
    if input.Strings[0] = '-' then begin
      if input.Strings[1] <> LEFT_PARENTHESIS then begin
          input.Strings[0] := '-' + input.Strings[1];
          input.Delete(1);
      end else begin
        input.Strings[0] := '-1';
        input.Insert(1, '*');
      end;
    end;

    // glowna petla
    for i := 0 to input.Count - 1 do begin // Pki zostay symbole do przeczytania wykonuj
      tmp := input.Strings[i]; // Przeczytaj symbol.
      if IsNumber2(tmp) or FConstants.IsMember2(tmp) then // Jeli symbol jest liczb
        output.Push(tmp) // dodaj go do kolejki wyjcie.
      else if FFunctions.IsMember2(tmp) then // Jeli symbol jest funkcj
        stack.Push(tmp) // w go na stos
      else if FOperators.GetItem(tmp, sym) <> -1 then begin // Jeli symbol jest operatorem,
        o1 := sym; // o1, wtedy
        if not stack.IsEmpty then
          repeat // dopki na grze stosu znajduje si operator,
            if stack.IsEmpty then
              b := False
            else if stack.Peek = LEFT_PARENTHESIS then // (wyklucz lewy nawias)
              b := False
            else begin
              o2 := FOperators.Find(stack.Peek); // o2 taki, e:
                    // o1 lewostronnie czny i jego kolejno wykonywania jest mniejsza lub rwna kolejnoci wyk. o2, lub
              b := ((not o1.AssociativeRight) and (o1.Priority <= o2.Priority)) or
                    // o1 jest prawostronnie czny i jego kolejno wykonywania jest mniejsza od o2,
                    (o1.AssociativeRight and (o1.Priority < o2.Priority));
            end;
            if b then
              output.Push(stack.Pop); // zdejmij o2 ze stosu i do go do kolejki wyjciowej;

          until not b;
        stack.Push(tmp); // w o1 na stos operatorw.

      end
      else if tmp = LEFT_PARENTHESIS then // Jeeli symbol jest lewym nawiasem to
        stack.Push(tmp) // w go na stos.
      else if tmp = RIGHT_PARENTHESIS then begin // Jeeli symbol jest prawym nawiasem to
        while not stack.IsEmpty do begin
          tmp2 := stack.Pop; // zdejmuj operatory ze stosu
          if tmp2 <> LEFT_PARENTHESIS then
            output.Push(tmp2) // i dokadaj je do kolejki wyjcie,
          else // dopki symbol na grze stosu nie jest lewym nawiasem,
            Break; // kiedy dojdziesz do tego miejsca zdejmij lewy nawias ze stosu bez dokadania go do kolejki wyjcie.
        end;
        if not stack.IsEmpty then
          if FFunctions.IsMember(stack.Peek) then // Teraz, jeli najwyszy element na stosie jest funkcj,
            output.Push(stack.Pop); // take do go do kolejki wyjcie.
      end;
    end;
    // Jeli nie ma wicej symboli do przeczytania,
    while not stack.IsEmpty do begin// zdejmuj wszystkie symbole ze stosu (jeli jakie s)
      tmp := stack.Pop;
      if (tmp = LEFT_PARENTHESIS) or (tmp = RIGHT_PARENTHESIS) then
        raise Exception.Create(PARENTHESIS_ERROR)
      else
        output.Push(tmp); // i dodawaj je do kolejki wyjcia.
    end;

    output.Delimiter := #32;
    Result := output.DelimitedText; // zwrocenie wartosci funkcji
  finally
    // zwalnianie stosow
    input.Free;
    stack.Free;
    output.Free;
  end;
end;

// funkcja konwertuje ONP na format infix
function TEval.PostfixToInfix(Value: string): string;
var
  i: Integer;
  input, stack: TStringStack;
  tmp, a, b: string;
begin
  try
    // tworzenie stosow
    input := TStringStack.Create;
    stack := TStringStack.Create;

    // wrzucenie tokenow na stos
    input.Delimiter := #32;
    input.DelimitedText := Value;
    input.DeleteEmpty;

    //glowna petla
    for i := 0 to input.Count - 1 do begin
      tmp := input.Strings[i];
      if IsNumber2(tmp) or FConstants.IsMember2(tmp) then // jeli i-ty symbol jest liczb,
        stack.Push(tmp) // odloz ja na stos
      else if FFunctions.IsMember2(tmp) then begin // jeli i-ty symbol jest funkcj
        a := stack.Pop; // zdejmij ze stosu a
        stack.Push(tmp + LEFT_PARENTHESIS + a + RIGHT_PARENTHESIS); // wrzuc na stos funkcje(a)
      end else if FOperators.IsMember(tmp) then begin // jeli i-ty symbol jest operatorem
        if tmp = '@' then tmp := '*';
        a := stack.Pop; // zdejmij ze stosu b
        b := stack.Pop; // zdejmij ze stosu a
        stack.Push(LEFT_PARENTHESIS + b + tmp + a + RIGHT_PARENTHESIS); // wrzuc na stos (b operator a)
      end else // nieznany symbol, wymus blad
        raise Exception.Create(Format(SYMBOL_ERROR, [input.Strings[i]]));
    end;

    stack.Delimiter := #32;
    Result := stack.DelimitedText; // Zdejmij ze stosu wynik.
    if (Result[1] = LEFT_PARENTHESIS) and (Result[Length(Result)] = RIGHT_PARENTHESIS) then
      Result := Copy(Result, 2, Length(Result) - 2);

  finally
    // zwalnianie stosow
    input.Free;
    stack.Free;
  end;
end;
// ----------------- TEval ------------------ //

end.
